/*
    IONLevel.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <cassert>
#include <dlfcn.h>

#include "util/schedYield.h"
#include "util/getABSTimeSpec.h"
#include "IONLevel.h"

using namespace std;

#define LCD_LINES 16
#define LCD_COLS (State::MAX_PARTS*2)

const char IONLevel::lcdChar = '=';
const short IONLevel::RED = 1;//οä˰̣Ϥʤ
const short IONLevel::GREEN = 2;//COLOR_PAIRΰ˻Ȥ
const short IONLevel::YELLOW = 3;//줾줬㤦ʤok


IONLevel::IONLevel (State* state__, WidiController* widiController__,
		    Option* option__):
  IONCurses (state__, widiController__, option__),
  hasColor (false)
{
  resetVolumeInfo ();
}

IONLevel::~IONLevel ()
{
  //do nothing
}



void IONLevel::initialize () throw (Error)
{
#if 0
  if (!(initscr ()))
    {
      Error error ("ncurses:can not open screen");
      throw (error);
    }
#endif

  screen=getScreen ();
  lockCurses ();
  (void) set_term (screen);
  infoWindowLines=4;
  infoWindowCols=COLS;
  infoWindow=newwin (infoWindowLines,infoWindowCols,0,0);

  mainWindowLines=LCD_LINES+1;
  mainWindowCols=COLS;
  mainWindow=newwin (mainWindowLines,mainWindowCols,LINES-mainWindowLines,0);

  messageWindowLines=LINES-infoWindowLines-mainWindowLines;
  messageWindowCols=COLS;
  messageWindow=newwin (messageWindowLines,messageWindowCols,
			infoWindowLines,0);
  assert (infoWindow!=NULL);
  assert (mainWindow!=NULL);
  assert (messageWindow!=NULL);
  start_color ();
  hasColor=false;
  if (has_colors ())
    {
      hasColor=true;
      init_pair (RED,   COLOR_RED,   COLOR_BLACK);
      //init_pair (RED,   COLOR_MAGENTA,   COLOR_BLACK);
      //init_pair (RED,   COLOR_CYAN,   COLOR_BLACK);
      //init_pair (GREEN, COLOR_GREEN, COLOR_BLACK);
      init_pair (GREEN, COLOR_CYAN, COLOR_BLACK);
      //init_pair (GREEN, COLOR_BLUE, COLOR_BLACK);
      init_pair (YELLOW,COLOR_YELLOW,COLOR_BLACK);
    }
  unlockCurses ();
  for (int i=0;i<State::MAX_PARTS/2;i++)
    {
      mvwprintw (mainWindow,LCD_LINES,i*4,"%02d",i*2+1);
      wattron (mainWindow,A_BOLD);
      mvwprintw (mainWindow,LCD_LINES,i*4+2,"%02d",i*2+2);
      wattroff (mainWindow,A_BOLD);
    }
  MainIO::initialize ();
}

void IONLevel::resetVolumeInfo ()
{
  for (int i=0;i<State::MAX_PARTS;i++)
    volumeInfo[i].reset ();
}

void IONLevel::reset ()
{
  resetVolumeInfo ();
}


/////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////ɽ
/////////////////////////////////////////////////////////////////////
void IONLevel::updateMainWindow ()
{
  showLevel ();

#define DISPLAY_TIME ((double) cacheState.sequencerStatus.displayTime*cacheState.songStatus.rate*720/100/15)
  //displayTime1դ7.2/15secondʤΤǡ
  
    if (cacheState.sequencerStatus.pageMachineTicks
	<=cacheState.sequencerStatus.dotMachineTicks
	&&
	cacheState.sequencerStatus.changedPage==0
	&&
        cacheState.songStatus.machineTicks-
        cacheState.sequencerStatus.dotMachineTicks<DISPLAY_TIME
	)
      showLCD (0);
    else if (cacheState.songStatus.machineTicks-
	     cacheState.sequencerStatus.pageMachineTicks<DISPLAY_TIME)
      showLCD (cacheState.sequencerStatus.displayPage);

#undef DISPLAY_TIME 
}

void IONLevel::showLevel ()
{
  char lcd[LCD_LINES][LCD_COLS];
  memset (lcd,' ',sizeof (lcd));
  
  for (int part=0;part<State::MAX_PARTS;part++)
    {
      //ޤvelocity׻
#if 1
      const uchar maxVelocity=cacheState.part[part].getMaxVelocity (cacheState.songStatus);
      const uint32 maxVelocityLevel=(uint32)
	(((double)cacheState.sequencerStatus.masterVolume*
	  cacheState.part[part].channelVolume*
	  maxVelocity*
	  cacheState.part[part].expression*
	  (LCD_LINES+1))/16383/16383/127/16383);//note 16384=14bit
      const uchar noteOnVelocity=cacheState.part[part].getNoteOnVelocity (cacheState.songStatus);
      const uint32 noteOnVelocityLevel=(uint32)
	(((double)cacheState.sequencerStatus.masterVolume*
	  cacheState.part[part].channelVolume*
	  noteOnVelocity*
	  cacheState.part[part].expression*
	  (LCD_LINES+1))*90/100/16383/16383/127/16383);//note 16384=14bit
      //90/100sustinʬ72˿̣Ϥʤ

      //ɤĹˤʤ뤫׻
      uint32 velocityLevel=maxVelocityLevel;
      if (noteOnVelocityLevel>maxVelocityLevel)
	velocityLevel=noteOnVelocityLevel;
      
#else
      const uint velocity=cacheState.part[part].getMaxVelocity (cacheState.songStatus);
      uint32 velocityLevel=(uint32)
	(((double)cacheState.sequencerStatus.masterVolume*
	  cacheState.part[part].partVolume*
	  velocity*
	  cacheState.part[part].expression*
	  (LCD_LINES+1))/16383/16383/127/16383);//note 16384=14bit
#endif

      //ʬȲäƤʬΰ֤η׻
#define DECREASING_TIME_MAX_DELAY (cacheState.songStatus.rate*29/100) //0.29sec
#define DECREASING_TIME_MAX (cacheState.songStatus.rate*13/100) //0.13sec
#define DECREASING_RATE_MAX 1
#define DECREASING_TIME_NOW (cacheState.songStatus.rate*4/100) // 0.04sec
#define DECREASING_RATE_NOW 1
      while (cacheState.songStatus.machineTicks-
	     volumeInfo[part].maxMachineTicks
	     >=DECREASING_TIME_MAX)
	{
	  volumeInfo[part].maxMachineTicks+=DECREASING_TIME_MAX;
	  if (volumeInfo[part].maxLevel<=DECREASING_RATE_MAX)
	    {
	      volumeInfo[part].maxLevel=0;
	      break;
	    }
	  volumeInfo[part].maxLevel-=DECREASING_RATE_MAX;
	}
      if (velocityLevel>=volumeInfo[part].maxLevel)
	{
	  volumeInfo[part].maxLevel=velocityLevel;
	  volumeInfo[part].maxMachineTicks=
	    //cacheState.part[part].lastMachineTicks+DECREASING_TIME_MAX_DELAY;
	    cacheState.songStatus.machineTicks+DECREASING_TIME_MAX_DELAY;
	}

      while (cacheState.songStatus.machineTicks-
	     volumeInfo[part].nowMachineTicks
	     >=DECREASING_TIME_NOW)

	{
	  volumeInfo[part].nowMachineTicks+=DECREASING_TIME_NOW;
	  if (volumeInfo[part].nowLevel<=DECREASING_RATE_NOW)
	    {
	      volumeInfo[part].nowLevel=0;
	      break;
	    }
	  volumeInfo[part].nowLevel-=DECREASING_RATE_NOW;
	}
      if (velocityLevel>=volumeInfo[part].nowLevel)
	{
	  volumeInfo[part].nowLevel=velocityLevel;
	  volumeInfo[part].nowMachineTicks=
	    //cacheState.part[part].lastMachineTicks;
	    cacheState.songStatus.machineTicks;
	}

      //bar('*')Τι
      assert (velocityLevel<=LCD_LINES);
      assert (volumeInfo[part].nowLevel<=LCD_LINES);
      assert (volumeInfo[part].maxLevel<=LCD_LINES);
      //for (int i=1;i<velocityLevel;i++)
      for (int i=1;i<volumeInfo[part].nowLevel;i++)
	{
	  lcd[LCD_LINES-i][part*2  ]=lcdChar;
	  lcd[LCD_LINES-i][part*2+1]=lcdChar;
	}
      if (volumeInfo[part].maxLevel>0)
	{
	  lcd[LCD_LINES-(uint) volumeInfo[part].maxLevel][part*2  ]=lcdChar;
	  lcd[LCD_LINES-(uint) volumeInfo[part].maxLevel][part*2+1]=lcdChar;
	}
    }

  //bar('=')ɽ
  const int attr=A_BOLD;
  
  if (hasColor)
    wattron (mainWindow,COLOR_PAIR (RED));
  for (int i=0;i<1;i++)
    {
      wmove (mainWindow,i,0);
      if (!hasColor)
	{
	  for (int j=0;j<LCD_COLS/4;j++)
	    {
	      waddnstr (mainWindow,lcd[i]+j*4,2);
	      wattron (mainWindow,attr);
	      waddnstr (mainWindow,lcd[i]+j*4+2,2);
	      wattroff (mainWindow,attr);
	    }
	}
      else
	{
	  wattron (mainWindow,attr);
	  waddnstr (mainWindow,lcd[i],LCD_COLS);
	  wattroff (mainWindow,attr);
	}
    }
  if (hasColor)
    {
      wattroff (mainWindow,COLOR_PAIR (RED));
      wattron (mainWindow,COLOR_PAIR (YELLOW));
    }
  for (int i=1;i<3;i++)
    {
      wmove (mainWindow,i,0);
      if (!hasColor)
	{
	  for (int j=0;j<LCD_COLS/4;j++)
	    {
	      waddnstr (mainWindow,lcd[i]+j*4,2);
	      wattron (mainWindow,attr);
	      waddnstr (mainWindow,lcd[i]+j*4+2,2);
	      wattroff (mainWindow,attr);
	    }
	}
      else
	{
	  wattron (mainWindow,attr);
	  waddnstr (mainWindow,lcd[i],LCD_COLS);
	  wattroff (mainWindow,attr);
	}
    }
  if (hasColor)
    {
      wattroff (mainWindow,COLOR_PAIR (YELLOW));
      wattron (mainWindow,COLOR_PAIR (GREEN));
    }
  for (int i=3;i<LCD_LINES;i++)
    {
      wmove (mainWindow,i,0);
      for (int j=0;j<LCD_COLS/4;j++)
	{
	  waddnstr (mainWindow,lcd[i]+j*4,2);
	  wattron (mainWindow,attr);
	  waddnstr (mainWindow,lcd[i]+j*4+2,2);
	  wattroff (mainWindow,attr);
	}
    }
  if (hasColor)
    {
      wattroff (mainWindow,COLOR_PAIR (GREEN));
    }
}



inline static void setLCD
(const int page,const uchar dotData[SequencerStatus::PAGE_NUM][SequencerStatus::DOT_SIZE],
 char lcd[LCD_LINES][LCD_COLS/2],const int dataOffset,const int columnOffset,
 char lcdChar)
{
  for (int i=0;i<16;i++)
    {
      for (int j=0;j<5;j++)
	{
	  if (((dotData[page][i+dataOffset])&(0x10 >> j))!=0)
	    {
	      lcd[i][(j+columnOffset)*2  ]=lcdChar;
	      lcd[i][(j+columnOffset)*2+1]=lcdChar;
	    }
	}
    }
}

void IONLevel::showLCD (const int page)
{
  char lcd[LCD_LINES][LCD_COLS/2];
  memset (lcd,' ',sizeof (lcd));

  ///lcdι
  //dotData 00-15
  setLCD (page,cacheState.sequencerStatus.dotData,lcd,0,0,lcdChar);
  //dotData 16-31
  setLCD (page,cacheState.sequencerStatus.dotData,lcd,16,5,lcdChar);
  //dotData 16-31
  setLCD (page,cacheState.sequencerStatus.dotData,lcd,32,10,lcdChar);
  //dotData 48-63
  for (int i=0;i<16;i++)
    {
      if ((cacheState.sequencerStatus.dotData[page][i+48]&0x10))
	{
	  lcd[i][15*2  ]=lcdChar;
	  lcd[i][15*2+1]=lcdChar;
	}
    }

  //ºݤɽ
  const int attr=A_BOLD;
  wattron (mainWindow,attr);
  for (int i=0;i<LCD_LINES;i++)
    mvwaddnstr (mainWindow,i,LCD_COLS/2,lcd[i],LCD_COLS/2);
  wattroff (mainWindow,attr);
}


void IONLevel::playLoop ()
{
  uint32 lastTicks = 0;
  uint timedOutCount = 0;//Ϣ³timed out?
  for (;;)
    {
#define TIMEDOUT 100000000L //100000000nsec=0.1sec
      //Ѳ̵ƤTIMEDOUTȤɽ򹹿
      timespec timeSpec;
      getABSTimeSpec (&timeSpec, 0, TIMEDOUT);

      int ret = 0;
      //֤ѲԤ
      state->lock ();
      while ((lastTicks == state->songStatus.ticks) &&
	     (state->songStatus.state != SongStatus::DONE) &&
	     (ret != ETIMEDOUT))
	ret=state->timedWaitChange (&timeSpec);
      state->getState (&cacheState);
      state->unlock ();

      if (ret == ETIMEDOUT && lastTicks == cacheState.songStatus.ticks)
	{
	  timedOutCount++;
	  //10ǳäƤΤTIMEDOUT=0.1 sec,ratehzñ̤ʤΤǡ
	  cacheState.songStatus.machineTicks +=
	    timedOutCount*cacheState.songStatus.rate/10;
	  cacheState.songStatus.logicalTicks +=
	    timedOutCount*SongStatus::SONG_CTRL_RATE/10;
	  cacheState.songStatus.ticks +=
	    timedOutCount*cacheState.songStatus.getDivision ()/
	    cacheState.songStatus.getTempo ()/10;
	}
      else
	{
	  timedOutCount = 0;
	}
      lastTicks = cacheState.songStatus.ticks;
      
      //Ūʽ
      update ();
      checkPause ();
      if (cacheState.songStatus.state == SongStatus::DONE)
	break;
      schedYield ();
    }
  resetVolumeInfo ();
#undef TIMEDOUT
}


#undef LCD_LINES
#undef LCD_COLS
#undef DECREASING_TIME_MAX_DELAY 
#undef DECREASING_TIME_MAX
#undef DECREASING_RATE_MAX
#undef DECREASING_TIME_NOW
#undef DECREASING_RATE_NOW



////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////for dynamic loading
////////////////////////////////////////////////////////////////////////////
#include "IO/InfoIO.h"
static IO* newIONLevel (State* state, WidiController* widiController,
			Option* option)
{
  return new IONLevel (state, widiController, option);
}

static MainIO* newMainIONLevel (State* state, WidiController* widiController,
				Option* option)
{
  return new IONLevel (state, widiController, option);
}

InfoIO infoIONLevel=
{
  newIONLevel,
  newMainIONLevel,
  false
};
