#include "SequencerSerial.h"
#include "Sequencer/MIDI.h"

SequencerSerial::SequencerSerial (State* globalState,Option* option__):
  Sequencer (globalState, option__),
  timedSerial (new TimedSerial ())
{
}

SequencerSerial::~SequencerSerial ()
{
  delete timedSerial;
}



uchar SequencerSerial::checkChannel (uchar channel)
{
  if (channel < MIDI::MAX_CHANNEL)
    {
      if(cacheState.sequencerStatus.port != SequencerStatus::PORT_A)
	portChange (SequencerStatus::PORT_A);
      return channel;
    }
  else if (MIDI::MAX_CHANNEL <= channel &&
	   channel < MIDI::MAX_CHANNEL*2)
    {
      if (cacheState.sequencerStatus.port != SequencerStatus::PORT_B)
	portChange (SequencerStatus::PORT_B);
      return channel-MIDI::MAX_CHANNEL;
    }

  return channel;
}

void SequencerSerial::sendAllChannel (uchar data[3])
{
  portChange (SequencerStatus::PORT_A);
  for (int channel = 0;
       channel < MIDI::MAX_CHANNEL;
       channel++)
    {
      timedSerial->write (data, 3);
      data[0]++;
    }
  portChange (SequencerStatus::PORT_B);
  data[0] -= MIDI::MAX_CHANNEL;
  for (int channel = 0;
       channel < MIDI::MAX_CHANNEL;
       channel++)
    {
      timedSerial->write (data, 3);
      data[0]++;
    }
  portChange (SequencerStatus::PORT_A);
}


//////////////////////////////////////////////////////////////////////

void SequencerSerial::dumpBuf ()
{
  traceDumpBuf ();
}

void SequencerSerial::allNoteOffSoundOff ()
{
  ///////////////////////////////all note off
  uchar noteOff[3] = {0x90,0x40,0x00};
  sendAllChannel (noteOff);
  //////////////////////////////all sound off
  uchar soundOff[3] = {0xB0,0x78,0x00};
  sendAllChannel (soundOff);

  for (int i = 0; i < State::MAX_PARTS; i++)
    {
      for (int j = 0; j < Part::MAX_NOTES; j++)
        {
          traceNoteOff (cacheState.part[i], j);
        }
    }
  dumpBuf ();
  sync ();

}

void SequencerSerial::reset ()
{
  allNoteOffSoundOff ();
  
  ////////////////////////////reset all controllers
  uchar reset[3] = {0xB0,0x79,0x00};
  sendAllChannel (reset);
  sync ();
  
  traceReset ();
}


//midi event¹Ԥ롣
void SequencerSerial::noteOff (uchar channel,uchar note,uchar velocity)
{
  traceNoteOff (channel,note,velocity);
  channel = checkChannel (channel);
  if (channel < MIDI::MAX_CHANNEL)
    {
      uchar buffer[3];
      buffer[0] = MIDI::NOTE_OFF | channel;
      buffer[1] = note;
      buffer[2] = velocity;
      timedSerial->write (buffer, 3);
    }
}

void SequencerSerial::noteOn (uchar channel,uchar note,uchar velocity)
{
  traceNoteOn (channel,note,velocity);
  channel = checkChannel (channel);
  if (isPronouce (channel, velocity))
    {
      if (channel < MIDI::MAX_CHANNEL)
	{
	  uchar buffer[3];
	  buffer[0] = MIDI::NOTE_ON | channel;
	  buffer[1] = note;
	  buffer[2] = velocity;
	  timedSerial->write (buffer, 3);
	}
    }
}

void SequencerSerial::polyphonicKeyPressure (uchar channel,
					     uchar note,uchar keyPressure)
{
  tracePolyphonicKeyPressure (channel,note,keyPressure);
  channel = checkChannel (channel);
  if (channel < MIDI::MAX_CHANNEL)
    {
      uchar buffer[3];
      buffer[0] = MIDI::KEY_PRESSURE | channel;
      buffer[1] = note;
      buffer[2] = keyPressure;
      timedSerial->write (buffer, 3);
    }
}

void SequencerSerial::controlChange (uchar channel,uchar control,uchar value)
{
  traceControlChange (channel,control,value);
  channel = checkChannel (channel);
  if (channel < MIDI::MAX_CHANNEL)
    {
      uchar buffer[3];
      buffer[0] = MIDI::CONTROL_CHANGE | channel;
      buffer[1] = control;
      buffer[2] = value;
      timedSerial->write (buffer, 3);
    }
}

void SequencerSerial::programChange (uchar channel,uchar  program)
{
  traceProgramChange (channel,program);
  channel = checkChannel (channel);
  if (channel < MIDI::MAX_CHANNEL)
    {
      uchar buffer[2];
      buffer[0] = MIDI::PROGRAM_CHANGE | channel;
      buffer[1] = program;
      timedSerial->write (buffer, 2);
    }
}

void SequencerSerial::channelPressure (uchar channel,uchar channelPressure)
{
  traceChannelPressure (channel,channelPressure);
  channel = checkChannel (channel);
  if (channel < MIDI::MAX_CHANNEL)
    {
      uchar buffer[2];
      buffer[0] = MIDI::CHANNEL_PRESSURE | channel;
      buffer[1] = channelPressure;
      timedSerial->write (buffer, 2);
    }
}

void SequencerSerial::pitchBendChange (uchar channel,uchar lsb,uchar msb)
{
  tracePitchBendChange (channel,lsb,msb);
  channel = checkChannel (channel);
  if (channel < MIDI::MAX_CHANNEL)
    {
      uchar buffer[3];
      buffer[0] = MIDI::PITCH_BEND | channel;
      buffer[1] = lsb;
      buffer[2] = msb;
      timedSerial->write (buffer, 3);
    }
}

void SequencerSerial::exclusive (const uchar* udata,
				 uint32 size, int channel = -1)
{
  traceExclusive (udata,size,channel);
  timedSerial->write (udata, size);
}


void SequencerSerial::portChange (int port)
{
  //0xF5 pp
  //pp<=1 port A,pp>=2 port B
  tracePortChange (port);
  uchar buffer[2];
  buffer[0] = 0xF5;
  buffer[1] = static_cast<uchar> (port+1);
  timedSerial->write (buffer, 2);
}

int SequencerSerial::getCtrlRate () throw (Error)
{
  return TimedSerial::CTRL_RATE;
}

void SequencerSerial::waitTimer (ulong ticks)
{
  timedSerial->waitTimer (ticks);
}

void SequencerSerial::deltaTime (ulong ticks)
{
  timedSerial->deltaTime (ticks);
}

void SequencerSerial::startTimer ()
{
  timedSerial->startTimer ();
}

void SequencerSerial::stopTimer ()
{
  timedSerial->stopTimer ();
}

void SequencerSerial::sync ()
{
  timedSerial->sync ();
}



////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////for dynamic loading
///////////////////////////////////////////////////////////newSequencer<T>
//////////////////////////////////////////////////////////infoSequencer<T>
////////////////////////////////////////////////////////////////////////////
#include "Sequencer/InfoSequencer.h"

static Sequencer* newSequencerSerial (State* state_, Option* option_)
{
  return new SequencerSerial (state_, option_);
}

InfoSequencer infoSequencerSerial =
{
  newSequencerSerial
};

