/*
    SongCommand.h
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */

#ifndef SongCommand_H
#define SongCommand_H

#include <pthread.h>

#include "util/mytype.h"
#include "util/Mutex.h"
#include "util/CondVar.h"
#include "Sequencer/Sequencer.h"

class SongCommand
{
public:
  enum Command
  {
    //to be handled by class Song
    NONE,
    SET, //$B1iAU0LCV$rJQ99(B
    //$B1iAU0LCV$r>/$7?J$a$k!#(B
    //($B$I$l$@$1?J$a$k$+$O1iAU$7$F$$$k(BSong$B$N(Bsub class$B<!Bh!#(B)
    FORWARD,
    //$B1iAU0LCV$r>/$7La$9!#(B
    //($B$I$l$@$1?J$a$k$+$O1iAU$7$F$$$k(BSong$B$N(Bsub class$B<!Bh!#(B)
    BACK,
    PAUSE,//$B1iAU$r0l;~Dd;_!#(B
    CONTINUE,//$B1iAU$r:F3+(B
    REWIND,//$B6J$N:G=i$KLa$k!#(B
    ABORT,//$B6J1iAU$rCf;_$9$k!#(B

    SET_SKEW,//skew$B$rJQ99$9$k!#(B(skew$B$O%F%s%]$r$:$i$9$N$K;H$&!#(B1.0$B$,DL>o:F@8!#(B)
    
    //midi port$B$K%G!<%?$r=PNO$9$k!#(B
    MIDI_PORT_NOTE_ON,
    MIDI_PORT_NOTE_OFF,
    MIDI_PORT_POLYPHONIC_KEY_PRESSURE,
    //MSB,LSB$B$rF1;~$K;XDj$9$kI,MW$,$"$k!#(B
    //RPN,NRPN$B$N>uBV$NA`:nA0$KLa$9$H$$$&;v$r$7$J$$$N$G!"(B
    //RPN,NRPN$B$rA`:n$7$?$$>l9g$O(BMIDI_PORT_PRN,MIDI_PORT_NPRN,
    //$B$r;H$&$3$H!#(B
    MIDI_PORT_CONTROL_CHANGE,
    MIDI_PORT_RPN,
    MIDI_PORT_NRPN,
    MIDI_PORT_PROGRAM_CHANGE,//MSB,LSB$BF1;~$K;XDj$9$kI,MW$,$"$k(B???
    MIDI_PORT_CHANNEL_PRESSURE,
    MIDI_PORT_PITCH_BEND_CHANGE,
    MIDI_PORT_SYS_EX,

    SET_MUTE,//$B;XDj$N(Bport,channel$B$r(Bmute$B$r$9$k!"$7$J$$$r;XDj$9$k!#(B
    TOGGLE_MUTE,//$B;XDj$N(Bport,channel$B$r(Bmute$B$r$9$k!"$7$J$$$r(Btoggle$B$9$k!#(B
  };
private:
  //////////////////////////////////////////for syncrrnous
  Mutex commandLock;
  CondVar commandCond;
  Command command;

  //$B1iAU0LCV;XDjMQ!#C10L$O(Blogical ticks
  ulong position;
  //skew$B;XDjMQ(B
  double skew;

  //MIDI_PORT_NOTE_ON,
  //MIDI_PORT_NOTE_OFF,
  //MIDI_PORT_POLYPHONIC_KEY_PRESSURE,
  //MIDI_PORT_CONTROL_CHANGE,
  //MIDI_PORT_PRN,
  //MIDI_PORT_NPRN,
  //MIDI_PORT_PROGRAM_CHANGE,
  //MIDI_PORT_CHANNEL_PRESSURE,
  //MIDI_PORT_PITCH_BEND_CHANGE,
  //$B$r;XDj$9$k$?$a$N%G!<%?(B
  uchar midiParameter[7];

  //MIDI_PORT_SYS_EX
  //$B$r;XDj$9$k$?$a$N%G!<%?(B
  int sysExSize;
  uchar* sysExData;//midi port$B$K=PNO$9$k(Bevent data

  //SET_MUTE$B;XDjMQ(B
  int muteSize;
  Sequencer::Muting* muting;

public:
  SongCommand ();
  SongCommand (const SongCommand& songCommand_);
  ~SongCommand ();
  //$B0J2<$N$N(Bmethod$B$O8F$S=P$7B&$G(Block$B$7$F$*$/I,MW$,$"$k!#(B
  void reset ()
    {
      command = NONE;
      position = 0;
      skew = 1.0;
      memset (midiParameter, 0, sizeof (midiParameter));
      sysExSize = -1;
      delete[] sysExData;
      sysExData = NULL;
      muteSize = -1;
      delete[] muting;
      muting = NULL;
    };
  //$B0z?t$,ITMW$J(Bcommand$B$r%;%C%H$9$k!#(B
  void setCommand (Command command_)
    {
      //command$B$r>e=q$-$7$F$$$J$$$+(Bchek
      assert (command == NONE);
      //$B0z?t$r<h$i$J$$(Bcommand$B$+%A%'%C%/(B
      assert (command_ != NONE);
      assert (command_ != SET);
      assert (command_ != SET_SKEW);
      assert (command_ != MIDI_PORT_NOTE_ON);
      assert (command_ != MIDI_PORT_NOTE_OFF);
      assert (command_ != MIDI_PORT_POLYPHONIC_KEY_PRESSURE);
      assert (command_ != MIDI_PORT_CONTROL_CHANGE);
      assert (command_ != MIDI_PORT_RPN);
      assert (command_ != MIDI_PORT_NRPN);
      assert (command_ != MIDI_PORT_PROGRAM_CHANGE);
      assert (command_ != MIDI_PORT_CHANNEL_PRESSURE);
      assert (command_ != MIDI_PORT_PITCH_BEND_CHANGE);
      assert (command_ != MIDI_PORT_SYS_EX);
      assert (command_ != SET_MUTE);
      assert (command_ != TOGGLE_MUTE);
      command=command_;
    };
  //SET command$B$r(Bset$B$9$k!#C10L$O(Blogical ticks
  void setCommandSET (ulong position_)
    {
      assert (command == NONE);
      position = position_;
      command = SET;
    };
    //SET command$B$r(Bset$B$9$k!#(B
  void setCommandSkew (double skew_)
    {
      assert (command == NONE);
      skew = skew_;
      command=SET_SKEW;
    };
  void setCommandNoteOn (uchar channel, uchar note, uchar velocity)
    {
      assert (command  ==  NONE);
      //XXX $B0z?t%A%'%C%/(B
      midiParameter[0] = channel;
      midiParameter[1] = note;
      midiParameter[2] = velocity;
      command = MIDI_PORT_NOTE_ON;
    };
  void setCommandNoteOff (uchar channel, uchar note, uchar velocity)
    {
      assert (command == NONE);
      //XXX $B0z?t%A%'%C%/(B
      midiParameter[0] = channel;
      midiParameter[1] = note;
      midiParameter[2] = velocity;
      command = MIDI_PORT_NOTE_OFF;
    };
  void setCommandPolyphonicKeyPressure (uchar channel, uchar note, uchar keyPressure)
    {
      assert (command == NONE);
      //XXX $B0z?t%A%'%C%/(B
      midiParameter[0] = channel;
      midiParameter[1] = note;
      midiParameter[2] = keyPressure;
      command = MIDI_PORT_POLYPHONIC_KEY_PRESSURE;
    };
  void setCommandControlChange (uchar channel,
				uchar msbControl, uchar msbValue,
				uchar lsbControl, uchar lsbValue)
    {
      assert (command == NONE);
      //XXX$B0z?t%A%'%C%/(B
      midiParameter[0] = channel;
      midiParameter[1] = msbControl;
      midiParameter[2] = msbValue;
      midiParameter[3] = channel;
      midiParameter[4] = lsbControl;
      midiParameter[5] = lsbValue;
      command = MIDI_PORT_CONTROL_CHANGE;
    };
  void setCommandRPN (uchar channel, uchar rpnMSB, uchar rpnLSB,
		      uchar cc1, uchar cc1Value,
		      uchar cc2, uchar cc2Value)
    {
      assert (command == NONE);
      //XXX$B0z?t%A%'%C%/(B
      assert ((cc1 == 0x06 && cc2 == 0x26) ||
	      (cc1 == 0x60 && cc1Value == 0x00) ||
	      (cc1 == 0x61 && cc1Value == 0x00));
      midiParameter[0] = channel;
      midiParameter[1] = rpnMSB;
      midiParameter[2] = rpnLSB;
      midiParameter[3] = cc1;
      midiParameter[4] = cc1Value;
      midiParameter[5] = cc2;
      midiParameter[6] = cc2Value;
      command = MIDI_PORT_RPN;
    };
  void setCommandNRPN (uchar channel, uchar nrpnMSB, uchar nrpnLSB,
		       uchar cc1, uchar cc1Value,
		       uchar cc2, uchar cc2Value)
    {
      assert (command == NONE);
      //XXX$B0z?t%A%'%C%/(B
      assert ((cc1 == 0x06 && cc2 == 0x26) ||
	      (cc1 == 0x60 && cc1Value == 0x00) ||
	      (cc1 == 0x61 && cc1Value == 0x00));
      midiParameter[0] = channel;
      midiParameter[1] = nrpnMSB;
      midiParameter[2] = nrpnLSB;
      midiParameter[3] = cc1;
      midiParameter[4] = cc1Value;
      midiParameter[5] = cc2;
      midiParameter[6] = cc2Value;
      command = MIDI_PORT_NRPN;
    };
  void setCommandProgramChange (uchar channel, uchar program)
    {
      assert (command == NONE);
      //XXX $B0z?t%A%'%C%/(B
      midiParameter[0] = channel;
      midiParameter[1] = program;
      command = MIDI_PORT_PROGRAM_CHANGE;
    };
  void setCommandChannelPressure (uchar channel, uchar channelPressure)
    {
      assert (command == NONE);
      //XXX $B0z?t%A%'%C%/(B
      midiParameter[0] = channel;
      midiParameter[1] = channelPressure;
      command = MIDI_PORT_CHANNEL_PRESSURE;
    };
  void setCommandPitchBendChange (uchar channel, uchar lsb_, uchar msb_)
    {
      assert (command == NONE);
      //XXX $B0z?t%A%'%C%/(B
      midiParameter[0] = channel;
      midiParameter[1] = lsb_;
      midiParameter[2] = msb_;
      command = MIDI_PORT_PITCH_BEND_CHANGE;
    };
  void setCommandSysEx (const uchar* udata_, uint32 sysExSize_, int channel = -1)
    {
      assert (udata_ != NULL);
      assert (sysExSize_ > 0);
      assert (command == NONE);
      //XXX$B0z?t%A%'%C%/(B
      sysExSize = sysExSize_;
      delete[] sysExData;
      sysExData = new uchar[sysExSize];
      memcpy (sysExData, udata_, sysExSize);
      command = MIDI_PORT_SYS_EX;
    };

  //isToggle == false$B$N;~$O(Bmuting_$B$K;XDj$5$l$?CM$K(Bmute$B$r(Bset$B$9$k!#(B
  //isToggle == true$B$N;~$O(Bmute$B$N>uBV$r(Btoggle$B$9$k!#(B(muting_$B$K;XDj$5$l$?CM$OL5;k$5$l$k!#(B)
  void setCommandMute (const Sequencer::Muting* muting_, int muteSize_,
		       bool isToggle = false)
    {
      assert (muting_ != NULL);
      assert (muteSize_ > 0);
      if (isToggle)
	command = TOGGLE_MUTE;
      else
	command = SET_MUTE;
      muteSize = muteSize_;
      delete[] muting;
      muting = Sequencer::Muting::newCopy (muting_, muteSize);
    };
  void setCommandToggleMute (const Sequencer::Muting* muting_, int muteSize_)
    {setCommandMute (muting_, muteSize_, true);};

  //songCommand$B$r<h$j=P$9!#(B
  Command peekCommand () const {return command;};
  void getSongCommand (SongCommand* songCommand_)
    {
      songCommand_->reset ();
      songCommand_->command = command;
      songCommand_->position = position;
      songCommand_->skew = skew;
      memcpy (songCommand_->midiParameter, midiParameter,
	      sizeof (midiParameter));
      songCommand_->sysExSize = sysExSize;
      if (sysExSize > 0)
	{
	  songCommand_->sysExData = new uchar[sysExSize];
	  memcpy (songCommand_->sysExData, sysExData, sysExSize);
	}
      songCommand_->muteSize = muteSize;
      if (muteSize > 0)
	{
	  assert (muting != NULL);
	  songCommand_->muting = Sequencer::Muting::newCopy (muting, muteSize);
	}
      reset ();
    }
  ulong getPosition () const {assert (command  == SET);return position;};
  double getSkew () const {assert (command == SET_SKEW);return skew;};
  const uchar* getMidiParameter () const
    {
      assert (command == MIDI_PORT_NOTE_ON ||
	      command == MIDI_PORT_NOTE_OFF ||
	      command == MIDI_PORT_POLYPHONIC_KEY_PRESSURE ||
	      command == MIDI_PORT_CONTROL_CHANGE ||
	      command == MIDI_PORT_RPN ||
	      command == MIDI_PORT_NRPN ||
	      command == MIDI_PORT_PROGRAM_CHANGE ||
	      command == MIDI_PORT_CHANNEL_PRESSURE ||
	      command == MIDI_PORT_PITCH_BEND_CHANGE);
      return midiParameter;
    };
  int getSysExSize () const
    {assert (command == MIDI_PORT_SYS_EX);return sysExSize;};
  const uchar* getSysExData () const
    {assert (command == MIDI_PORT_SYS_EX);return sysExData;};
  int getMuteSize () const 
    {assert (command == SET_MUTE || command == TOGGLE_MUTE);return muteSize;};
  const Sequencer::Muting* getMuting () const
    {assert (command == SET_MUTE || command == TOGGLE_MUTE );return muting;};
  
  void lock () {commandLock.lock ();};
  int trylock () {return commandLock.trylock ();};
  void unlock () {commandLock.unlock ();};
  void broadcast (){commandCond.broadcast ();};
  void waitCond (){commandCond.wait (commandLock);};
  int timedWaitCond (const timespec* abstime)
    {return commandCond.timedWait (commandLock,abstime);};
};
#endif SongCommand_H
