/*
    State.h
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */
#ifndef State_H
#define State_H

#include <pthread.h>

#include "util/Error.h"
#include "util/Mutex.h"
#include "util/CondVar.h"
#include "Song/SongStatus.h"
#include "Sequencer/Part.h"
#include "Sequencer/SequencerStatus.h"

class State
{
public:
  static const int MAX_PARTS = 32;
private:
  //////////////////////////////////////////for syncronization
  Mutex mutexLock;
  CondVar hasBeenChanged;
public:
  State ();
  ~State ();
  //$BJQ?t$N=i4|2=!#(Block$B$O8F$S=P$9B&$G9T$&$3$H!#(B
  //songStatus$B$N(Bsong$B8GM-$N>pJs$O(Breset$B$7$J$$!#(B
  void reset () throw (Error);
  //$BJQ?t$N=i4|2=!#(Block$B$O8F$S=P$9B&$G9T$&$3$H!#(B
  //songStatus$B$N(Bsong$B8GM-$N>pJs$b(Breset$B$9$k!#(B
  void fullReset () throw (Error);
  //$BJQ?t$N=i4|2=!#(Block$B$O8F$S=P$9B&$G9T$&$3$H!#(B
  void partReset () throw (Error);

  void lock () {mutexLock.lock ();};
  int trylock () {return mutexLock.trylock ();};
  void unlock () {mutexLock.unlock ();};
  void broadcast (){hasBeenChanged.broadcast ();};
  void waitChange (){hasBeenChanged.wait (mutexLock);};
  int timedWaitChange (const timespec* abstime)
    {return hasBeenChanged.timedWait (mutexLock,abstime);};

  //$B$3$l$i$O8F$S=P$7B&$G%m%C%/$9$kI,MW$,$"$k!#(B
  inline void updateSongStatus (const SongStatus& midiSongStatus)
    {songStatus=midiSongStatus;};
  inline void updateSongStatus (const SongStatus* midiSongStatus)
    {songStatus=*midiSongStatus;};
  inline void updatePartStatus (const Part midiPart[MAX_PARTS])
    {
      for (int i=0;i<MAX_PARTS;i++)
	part[i]=midiPart[i];
    };
  inline void updateSequencerStatus (const SequencerStatus midiSequencerStatus)
    {sequencerStatus=midiSequencerStatus;};
  inline void getState (State* state_)
    {
      state_->songStatus=songStatus;
      for (int i=0;i<MAX_PARTS;i++)
	state_->part[i]=part[i];
      state_->sequencerStatus=sequencerStatus;
    };
  
public:
  ///////////////////////////////////////////////////Song status
  SongStatus songStatus;
  ///////////////////////////////////////////////Part status
  Part part[MAX_PARTS];
  /////////////////////////////////////////////////Sequencer status
  SequencerStatus sequencerStatus;
};
#endif //State_H
