/*
    DlHandler.h
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#ifndef DlHandler_H
#define DlHandler_H

#include <iostream>
#include <string>
#include <dlfcn.h>

#include "Widi/Option.h"

using namespace std;

class DlHandler
{
protected:
  Option* const option;
  
  void* handle;
public:
  DlHandler (Option* option_):option (option_), handle (NULL){/*do nothing*/};
  ~DlHandler (){close ();};
#ifdef RTLD_GLOBAL
  const void* open (const string& filename, int flag = RTLD_GLOBAL | RTLD_LAZY)
#else
  const void* open (const string& filename, int flag = RTLD_LAZY)//Does this work well?
#endif
    {
      assert (handle == NULL);
      handle = dlopen (filename.c_str (), flag);

      if (handle != NULL)
	return handle;//library found

      for (list<string>::const_iterator i =option->modulePath.begin ();
	   i != option->modulePath.end ();
	   i++)
	{//library not found.search direcotries
	  string absFilename;
	  absFilename = *i;
	  absFilename += "/";
	  absFilename += filename;

	  handle = dlopen (absFilename.c_str (), flag);
	  if (handle != NULL)
	    break;
	}
#if 0
      if (handle == NULL)
	{//library not found in the directoryies.assuming statically linked.
	  cerr << "DlHandler:can not load plug in=" << filename
	       << " in following dirctories.";
	  for (list<string>::const_iterator i = option->modulePath.begint ();
	       i != option->modulePath.end ();
	       i++)
	    cerr << ":" << *i;
	  cerr << "\nassuming statically linked" << endl;
	  handle = dlopen (NULL, flag);
	}
#endif
      if (handle == NULL)
	{//library not found at all.
	  cerr << "DlHandler:can not load plug in=" << filename
	       << " in following dirctories.";
	  for (list<string>::const_iterator i = option->modulePath.begin ();
	       i != option->modulePath.end ();
	       i++)
	    cerr << ":" << *i;
	  cerr << endl << dlerror () << endl;
	}
      return handle;
    };

  const void* openLibSo (const string& prefix, const string& name)
    {//"lib"+prefix+name+".so"$B$r(Bopen$B$9$k!#(B
      string libPrefixNameSo = "lib";
      libPrefixNameSo += prefix;
      libPrefixNameSo += name;
      libPrefixNameSo += ".so";

      const void* ret = open (libPrefixNameSo);
      return ret;
    };

  void* sym (string& symbol)
    {
      assert (handle != NULL);
      void* ret;
      ret = dlsym (handle, symbol.c_str ());
      if (ret == NULL)
	{
	  cerr << "DlHandler:can not resolve symbol=" << symbol << endl;
	  cerr << dlerror () << endl;
	}
      return ret;
    };

  void close ()
    {
      if (handle != NULL)
	{
	  dlclose (handle);
	  handle = NULL;
	}
    };

  static DlHandler* newOpenDlHandler (const string& prefix, const string& name,
				      Option* option)
    {
      assert (option != NULL);
      DlHandler* retHandler = new DlHandler (option);
      if (retHandler->openLibSo (prefix, name) == NULL)
	{
	  delete retHandler;
	  retHandler = NULL;
	}
      return retHandler;
    }
};
#endif //DlHandler_H
